package com.jigoku.backend;

import com.badlogic.gdx.graphics.Color;
import com.badlogic.gdx.graphics.g2d.SpriteBatch;
import com.badlogic.gdx.graphics.g2d.TextureRegion;
import com.badlogic.gdx.math.Intersector;
import com.badlogic.gdx.math.Rectangle;
import com.badlogic.gdx.scenes.scene2d.actors.Image;
import com.jigoku.objects.ObjectHandler;
import com.jigoku.projectiles.BulletHandler;

public abstract class AbstractObject extends Image {
	public static final String DefaultName = "Unnamed";
	
	//////////////////////////////////////////////////////////////////////////////////////////////
	// SPECIFIC ENUMS
	//////////////////////////////////////////////////////////////////////////////////////////////
	public enum CollisionType {
		Rectangle,
		Circle,
	}
	
	///////////////////////////////////////////////////////////////////////////////////////////////
	// INSTANCE VARIABLES
	///////////////////////////////////////////////////////////////////////////////////////////////
	public CollisionType collisionType;
	public Rectangle boundingBox;
	protected float stateTime;
	
	///////////////////////////////////////////////////////////////////////////////////////////////
	// CONSTRUCTORS
	///////////////////////////////////////////////////////////////////////////////////////////////
	public AbstractObject(AbstractObject o) {
		super(o.name);
		
		//Actor
		this.x = o.x;
		this.y = o.y;
		this.width = o.width;
		this.height = o.height;
		this.originX = o.originX;
		this.originY = o.originY;
		this.scaleX = o.scaleX;
		this.scaleY = o.scaleY;
		this.rotation = o.rotation;
		this.color.set(o.color);
		
		//Image
		this.region = o.region;
		
		//AbstractObject
		this.collisionType = o.collisionType;
		this.boundingBox = o.boundingBox;
		this.stateTime = o.stateTime;
	}
	
	public AbstractObject() {
		this(AbstractObject.DefaultName);
	}
	
	public AbstractObject(String name) {
		super(name);
		this.collisionType = CollisionType.Rectangle;
		this.boundingBox = null;
		this.stateTime = 0.0f;
	}
	
	///////////////////////////////////////////////////////////////////////////////////////////////
	// ABSTRACT METHODS
	///////////////////////////////////////////////////////////////////////////////////////////////
	public abstract AbstractObject clone();
	public abstract void create();
	public abstract void destroy();
	public abstract void act(float dt);
	public abstract void endOfAnimation();
	public abstract void checkCollisionWithObject(AbstractObject other);
	public abstract void checkCollisionWithObjects(ObjectHandler handler);
	public abstract void checkCollisionWithProjectiles(BulletHandler handler);
	public abstract void collisionHasOccurredWithObject(AbstractObject other);
	
	///////////////////////////////////////////////////////////////////////////////////////////////
	// MUTATORS / SETTERS
	///////////////////////////////////////////////////////////////////////////////////////////////
	public void setRegion(TextureRegion region) {
		this.width = region.getRegionWidth();
		this.height = region.getRegionHeight();
		this.region = region;
	}
	
	public void setPosition(float x, float y) {
		this.x = x;
		this.y = y;
	}
	
	public void translate(float dx, float dy) {
		this.x += dx;
		this.y += dy;
	}
	
	
	public void setOrigin(float ox, float oy) {
		this.originX = ox;
		this.originY = oy;
	}
	
	public void setScale(float s) {
		setScale(s, s);
	}
	
	public void setScale(float sx, float sy) {
		this.scaleX = sx;
		this.scaleY = sy;
	}
	
	public void setColor(Color c) {
		this.color.set(c);
	}
	
	public void setColor(float r, float g, float b, float a) {
		this.color.set(r, g, b, a);
	}
	
	public void setColor(int r, int g, int b, int a) {
		this.color.set(r/255.0f, g/255.0f, b/255.0f, a/255.0f);
	}
	
	///////////////////////////////////////////////////////////////////////////////////////////////
	// ACCESSORS / GETTERS
	///////////////////////////////////////////////////////////////////////////////////////////////
	public void set(AbstractObject o) {
		//Actor
		this.x = o.x;
		this.y = o.y;
		this.width = o.width;
		this.height = o.height;
		this.originX = o.originX;
		this.originY = o.originY;
		this.scaleX = o.scaleX;
		this.scaleY = o.scaleY;
		this.rotation = o.rotation;
		this.color.set(o.color);
		
		//Image
		this.region = o.region;
		
		//AbstractObject
		this.collisionType = o.collisionType;
		if(o.boundingBox != null)
			this.boundingBox = new Rectangle(o.boundingBox);
		else
			this.boundingBox = null;
		this.stateTime = o.stateTime;
	}
	
	///////////////////////////////////////////////////////////////////////////////////////////////
	// PUBLIC METHODS
	///////////////////////////////////////////////////////////////////////////////////////////////
	public Rectangle worldBoundingBox() {
		if(boundingBox == null) {
			return new Rectangle(x, y, width, height);
		} else {
			return new Rectangle(x + boundingBox.getX(), y + boundingBox.getY(),
								boundingBox.getWidth(), boundingBox.getHeight());
		}
	}
	
	public void draw(SpriteBatch batch) {
		super.draw(batch, 1.0f);
	}
	
	///////////////////////////////////////////////////////////////////////////////////////////////
	// PUBLIC STATIC METHODS
	///////////////////////////////////////////////////////////////////////////////////////////////
	public static boolean collisionTest(AbstractObject first, AbstractObject second) {
		Rectangle 	a = first.worldBoundingBox(),
					b = second.worldBoundingBox();
		return (Intersector.overlapRectangles(a, b));
	}
}
