package com.jigoku.objects;

import com.badlogic.gdx.Gdx;
import com.badlogic.gdx.graphics.Texture.TextureWrap;
import com.badlogic.gdx.graphics.g2d.SpriteBatch;
import com.badlogic.gdx.graphics.g2d.TextureRegion;
import com.badlogic.gdx.math.Rectangle;
import com.jigoku.ApplicationOptions;
import com.jigoku.backend.AbstractObject;
import com.jigoku.backend.Director;

public class Beam extends Object {
	///////////////////////////////////////////////////////////////////////////////////////////////
	// STATIC LOAD VARIABLES
	///////////////////////////////////////////////////////////////////////////////////////////////
	public static final float cutoff = ApplicationOptions.APP_HEIGHT;
	public static final float paddingWidth = 1.0f;
	public static final float paddingHeight = 5.0f;
	
	///////////////////////////////////////////////////////////////////////////////////////////////
	// SPECIFIC OBJECT ENUMS
	///////////////////////////////////////////////////////////////////////////////////////////////
	public enum BeamState {
		NotFiring,					//Default State - when beam is not being fired		
		BeginFiring,				//Set State 	- Beam first begins firing
		StopFiring,					//Set State		- beam first stops firing
		Firing_MissingTarget,		//Access State	- Beam is firing and is missing all targets
		Firing_HittingTarget,		//Access State	- Beam is firing and is hitting a target
		Firing_Processing,			//Access State	- Beam is firing but is NOT off-screen
		EndFiring_Processing,		//Access State	- Beam has stopped firing but is NOT off-screen
		EndFiring_HittingTarget,	//Access State	- Beam has stopped firing but is hitting target
		EndFiring_MissingTarget,	//Access State	- Beam has stopped firing and is missing all targets.
	}
	
	///////////////////////////////////////////////////////////////////////////////////////////////
	// INSTANCE VARIABLES
	///////////////////////////////////////////////////////////////////////////////////////////////
	public float beamSpeed, totalTimeMissing, totalTimeHitting;
	protected boolean 	isFiring, 		// Is currently being fired
						wasFiring, 		// Was being fired (i.e. Not currently, but not off-screen)
						wasMissing;		// Was missing targets
	protected AbstractObject target;
	protected BeamState currentState;
	protected float sourceY;

	///////////////////////////////////////////////////////////////////////////////////////////////
	// CONSTRUCTORS
	///////////////////////////////////////////////////////////////////////////////////////////////
	public Beam(String name, TextureRegion beamTexture, float beamSpeed, Director d) {
		super(name, d);
		
		//Set Texture
		this.setRegion(beamTexture);
		this.region.getTexture().setWrap(TextureWrap.ClampToEdge, TextureWrap.Repeat);
		
		//Set Instance Variables
		this.beamSpeed = beamSpeed;
		this.totalTimeMissing = 0.0f;
		this.totalTimeHitting = 0.0f;
		this.setOrigin(width / 2, 0.0f);
		this.resetInstanceVariables();
		this.setState(BeamState.NotFiring);
		this.boundingBox = new Rectangle(0.0f,0.0f,0.0f,0.0f);
	}
	
	///////////////////////////////////////////////////////////////////////////////////////////////
	// ENUM STATE HANDLING
	///////////////////////////////////////////////////////////////////////////////////////////////
	
	private void setState(BeamState s) {
		if(currentState == s)
			return;
		switch(s) {
		case BeginFiring:
			resetInstanceVariables();
			isFiring = true;
			wasFiring = true;
			setState(BeamState.Firing_Processing);
			return;
		case StopFiring:
			switch(currentState) {
			case Firing_Processing:
				setState(BeamState.EndFiring_Processing);
				break;
			case Firing_MissingTarget:
				setState(BeamState.EndFiring_MissingTarget);
				break;
			case Firing_HittingTarget:
				setState(BeamState.EndFiring_HittingTarget);
				break;
			}
			isFiring = false;
			return;
		case Firing_Processing:
			break;
		case Firing_MissingTarget:
			wasMissing = true;
			break;
		case Firing_HittingTarget:
			wasMissing = false;
			break;
		case EndFiring_Processing:
			break;
		case EndFiring_HittingTarget:
			wasMissing = false;
			break;
		case EndFiring_MissingTarget:
			wasMissing = true;
			break;
		case NotFiring:
			resetInstanceVariables();
			break;
		default:
			break;
		}
		currentState = s;
		//Debug Purposes
		//details();
	}
	
	///////////////////////////////////////////////////////////////////////////////////////////////
	// METHODS
	///////////////////////////////////////////////////////////////////////////////////////////////
	public void activate() {
		if(!isFiring)
			setState(BeamState.BeginFiring);
	}
	
	public void deactivate() {
		if(isFiring)
			setState(BeamState.StopFiring);
	}
	
	///////////////////////////////////////////////////////////////////////////////////////////////
	// OVERRIDES
	///////////////////////////////////////////////////////////////////////////////////////////////
	@Override
	public void act(float dt) {
		
		//Move dynamically
		switch(currentState) {
		case BeginFiring:
		case Firing_Processing:
			moveBeamHeight(dt);
			if(beamMeetsCutoff())
				setState(BeamState.Firing_MissingTarget);
			break;
		case Firing_HittingTarget:
			totalTimeHitting += dt;
			setHeightToTarget();
			scrollBeamTexture(dt);
			checkIfStillHittingTarget();
			break;
		case Firing_MissingTarget:
			totalTimeMissing += dt;
			setHeightToCutoff();
			scrollBeamTexture(dt);
			break;
		case EndFiring_Processing:
			moveBeamOrigin(dt);
			if(beamMeetsCutoff())
				setState(BeamState.EndFiring_MissingTarget);
			break;
		case EndFiring_HittingTarget:
			totalTimeHitting += dt;
			setHeightToTarget();
			scrollBeamTexture(dt);
			moveBeamOrigin(dt);
			if(beamOriginMeetsTarget())
				setState(BeamState.NotFiring);
			else
				checkIfStillHittingTarget();
			break;
		case EndFiring_MissingTarget:
			totalTimeMissing += dt;
			setHeightToCutoff();
			scrollBeamTexture(dt);
			moveBeamOrigin(dt);
			if(beamOriginMeetsCutoff())
				setState(BeamState.NotFiring);
			break;
		case NotFiring:
		case StopFiring:
			if(wasMissing)
				setState(BeamState.EndFiring_MissingTarget);
			else if(wasFiring)
				setState(BeamState.EndFiring_HittingTarget);
			break;
		}
		
		boundingBox.set(-Beam.paddingWidth, offsetY, 
				width + Beam.paddingWidth, height + Beam.paddingHeight);
	}
	
	@Override
	public void draw(SpriteBatch batch) {
		float 	tx = x - originX + offsetX,
				ty = y - originY + offsetY;
		batch.draw(region.getTexture(), 						// Texture
				tx, ty,											// Destination (x, y)
				originX, originY, 								// Destination (originX, originY)
				width, height,									// Destination (width, height) 
				scaleX, scaleY,									// Destination (scaleX, scaleY)
				rotation,										// Rotation
				0, (int)sourceY,								// Source (x, y)
				(int)width, (int)height,						// Source (width, height)
				flipX, flipY									// Flip (x, y)
				);
	}
	
	@Override
	public void checkCollisionWithObject(AbstractObject other) {
		if(currentState == BeamState.NotFiring)
			return;
		if(Object.collisionTest(this, other)) {
			if(target == null || other.worldBoundingBox().getY() < target.worldBoundingBox().getY()) {
				setTarget(other);
			}
		}
	}
	
	///////////////////////////////////////////////////////////////////////////////////////////////
	// PRIVATE METHODS
	///////////////////////////////////////////////////////////////////////////////////////////////
	private void setHeightToTarget() {
		height = target.worldBoundingBox().getY() - y - offsetY;
	}
	
	private void setHeightToCutoff() {
		height = Beam.cutoff - y - offsetY;
	}
	
	private boolean beamOriginMeetsTarget() {
		if(offsetY >= target.worldBoundingBox().getY() - y) {
			return true;
		}
		return false;
	}
	
	private boolean beamOriginMeetsCutoff() {
		if(offsetY >= Beam.cutoff - y) {
			return true;
		}
		return false;
	}
	
	private boolean beamMeetsCutoff() {
		if(height >= Beam.cutoff - offsetY - y){
			height = Beam.cutoff - offsetY - y;
			return true;
		}
		return false;
	}
	
	private void moveBeamHeight(float dt) {
		height += beamSpeed * dt;
	}
	
	private void scrollBeamTexture(float dt) {
		sourceY += (beamSpeed * dt);
		if(sourceY > beamSpeed)
			sourceY -= beamSpeed;
	}
	
	private void moveBeamOrigin(float dt) {
		offsetY += beamSpeed * dt;
	}
	
	private void resetInstanceVariables() {
		offsetY = 0.0f;
		height = 0.0f;
		sourceY = 0.0f;
		isFiring = false;
		wasFiring = false;
		wasMissing = false;
		setTarget(null);
	}
	
	private void setTarget(AbstractObject t) {
		target = t;
		if(target == null) {
			if(isFiring) {
				if(wasMissing)
					setState(BeamState.Firing_MissingTarget);
				else
					setState(BeamState.Firing_Processing);
			} else if(wasFiring) {
				if(wasMissing)
					setState(BeamState.EndFiring_MissingTarget);
				else
					setState(BeamState.EndFiring_Processing);
			}
		} else {
			if(isFiring)
				setState(BeamState.Firing_HittingTarget);
			else if(wasFiring)
				setState(BeamState.EndFiring_HittingTarget);
		}
	}
	
	private void checkIfStillHittingTarget() {
		if(!AbstractObject.collisionTest(this, target)) {
			setTarget(null);
		}
	}
	
	///////////////////////////////////////////////////////////////////////////////////////////////
	// ACCESSOR METHODS
	///////////////////////////////////////////////////////////////////////////////////////////////
	public boolean getIsFiring() {
		return isFiring;
	}
	
	public boolean getWasFiring() {
		return wasFiring;
	}
	
	public boolean getWasMissing() {
		return wasMissing;
	}
	
	public AbstractObject getTarget() {
		return target;
	}
	
	public BeamState getCurrentState() {
		return currentState;
	}

	///////////////////////////////////////////////////////////////////////////////////////////////
	// DEBUG METHODS
	///////////////////////////////////////////////////////////////////////////////////////////////
	@SuppressWarnings("unused")
	private void details() {
		String state = null;
		switch(currentState) {
		case BeginFiring:
			state = "BeginFiring";
			break;
		case StopFiring:
			state = "StopFiring";
			break;
		case Firing_Processing:
			state = "Firing_Processing";
			break;
		case Firing_MissingTarget:
			state = "Firing_MissingTarget";
			break;
		case Firing_HittingTarget:
			state = "Firing_HittingTarget";
			break;
		case EndFiring_Processing:
			state = "EndFiring_Processing";
			break;
		case EndFiring_HittingTarget:
			state = "EndFiring_HittingTarget";
			break;
		case EndFiring_MissingTarget:
			state = "EndFiring_MissingTarget";
			break;
		case NotFiring:
			state = "NotFiring";
			break;
		default:
			state = "Undefined";
			break;
		}
		Gdx.app.log("PlayerBullet state change:", state);
	}
}