package com.jigoku.projectiles;

import com.badlogic.gdx.Gdx;
import com.badlogic.gdx.graphics.g2d.SpriteBatch;
import com.badlogic.gdx.graphics.g2d.TextureAtlas;
import com.badlogic.gdx.graphics.g2d.TextureRegion;
import com.badlogic.gdx.math.Rectangle;
import com.jigoku.backend.Helper;
import com.jigoku.objects.Object;
import com.jigoku.objects.ObjectHandler;

public class BulletHandler {
	private static TextureAtlas atlas = null;
	public static float PaddingWidth = 50.0f;
	public static float PaddingHeight = 50.0f;
	public static TextureRegion standardBullet, largeBullet, trailBullet;
	
	///////////////////////////////////////////////////////////////////////////////////////////////
	// INSTANCE VARIABLES
	///////////////////////////////////////////////////////////////////////////////////////////////
	private boolean isActing;
	private boolean deactivateFlag;
	private int maxProjectiles;
	private int activeProjectiles;
	private Projectile[] projectiles;
	public Rectangle activeZone;
	
	///////////////////////////////////////////////////////////////////////////////////////////////
	// CONSTRUCTORS
	///////////////////////////////////////////////////////////////////////////////////////////////
	public BulletHandler(int maxProjectiles, Rectangle zone) {
		this.maxProjectiles = maxProjectiles;
		this.activeProjectiles = 0;
		this.projectiles = new Projectile[maxProjectiles];
		for(int i = 0; i < maxProjectiles; i++)
			projectiles[i] = new Projectile(this, i);
		this.activeZone = zone;
		this.isActing = false;
		this.deactivateFlag = false;
	}
	
	public BulletHandler(int maxProjectiles) {
		this(maxProjectiles, new Rectangle(-PaddingWidth, -PaddingHeight, Gdx.graphics.getWidth() + PaddingWidth, Gdx.graphics.getHeight() + PaddingHeight));
	}
	
	///////////////////////////////////////////////////////////////////////////////////////////////
	// ACCESSORS / GETTERS
	///////////////////////////////////////////////////////////////////////////////////////////////
	public int getMaxProjectiles() {
		return maxProjectiles;
	}
	
	public int getActiveProjectiles() {
		return activeProjectiles;
	}
	
	public Projectile getProjectile(int index) {
		if(index >= maxProjectiles || index < 0)
			return null;
		return projectiles[index];
	}
	
	///////////////////////////////////////////////////////////////////////////////////////////////
	// PUBLIC METHODS
	///////////////////////////////////////////////////////////////////////////////////////////////
	
	/*
	 * BULLET CREATION METHODS
	 */
	public Projectile createProjectile() {
		if(activeProjectiles >= maxProjectiles)
			return null;
		projectiles[activeProjectiles].set(Projectile.Default);
		projectiles[activeProjectiles].activate();
		activeProjectiles++;
		return projectiles[activeProjectiles - 1];
	}
	
	public Projectile createProjectile(float x, float y) {
		Projectile p = createProjectile();
		if(p != null)
			p.setPosition(x, y);
		return p;
	}
	
	public Projectile createProjectile(Projectile template) {
		Projectile p = createProjectile();
		if(p != null) {
			p.set(template);
			//Needs reactivated, setting returns Projectile to an Inactive state.
			p.activate();
		}
		return p;
	}
	
	///
	
	public void removeProjectile(int index) {
		int lastIndex = activeProjectiles - 1;
		
		Projectile swap = projectiles[index];
		projectiles[index] = projectiles[lastIndex];
		projectiles[lastIndex] = swap;
		
		//Index == Still alive | lastIndex == Dead
		projectiles[index].index = index;
		projectiles[lastIndex].index = lastIndex;
		activeProjectiles--;
		
		if(isActing)
			deactivateFlag = true;
	}
	
	public void act(float dt) {
		isActing = true;
		deactivateFlag = false;
		for(int i = 0; i < activeProjectiles; i++) {
			projectiles[i].act(dt);
			if(deactivateFlag) {
				i--;
				deactivateFlag = false;
				continue;
			}
			if(!Helper.RectangleContainsPoint(activeZone, projectiles[i].x, projectiles[i].y)) {
				removeProjectile(i);
				deactivateFlag = false;
				i--;
				continue;
			}
		}
		isActing = false;
		this.debug();
	}
	
	public void draw(SpriteBatch batch) {
		for(int i = 0; i < activeProjectiles; i++)
			projectiles[i].draw(batch);
	}
	
	public void checkCollisionWithObject(Object object) {
		for(int i = 0; i < activeProjectiles; i++) {
			if(Object.collisionTest(projectiles[i], object)) {
				projectiles[i].collisionHasOccurredWithObject(object);
				object.collisionHasOccurredWithObject(projectiles[i]);
				if(deactivateFlag) {
					i--;
					deactivateFlag = false;
					continue;
				}
			}
		}
	}
	
	public void checkCollisionWithObjects(ObjectHandler handler) {
		for(Object o : handler.objects)
			checkCollisionWithObject(o);
	}
	
	///////////////////////////////////////////////////////////////////////////////////////////////
	// STATIC METHODS
	///////////////////////////////////////////////////////////////////////////////////////////////
	public static void load() {
		if(atlas == null) {
			atlas = new TextureAtlas("data/sprites/bullets.txt");
			standardBullet = atlas.findRegion("standardBullet");
			trailBullet = atlas.findRegion("trailBullet");
			largeBullet = atlas.findRegion("largeBullet");
		}
	}
	
	public static void unload() {
	}
	
	///////////////////////////////////////////////////////////////////////////////////////////////
	// DEBUG METHODS
	///////////////////////////////////////////////////////////////////////////////////////////////
	public void debug() {
		Gdx.app.log("BulletHandler", String.format("Active: %d Max: %d",activeProjectiles, maxProjectiles));
	}
}
