package com.jigoku.projectiles;

import com.badlogic.gdx.graphics.g2d.TextureRegion;
import com.badlogic.gdx.math.MathUtils;
import com.jigoku.backend.AbstractObject;
import com.jigoku.backend.Helper;
import com.jigoku.objects.Object;
import com.jigoku.objects.ObjectHandler;
import com.jigoku.projectiles.Splitter.SplitterType;

public class Projectile extends AbstractObject {
	public static final String DefaultName = "Projectile";
	public static final int SigFig = 10000;
	public static final Projectile Default = new Projectile();
	
	public enum ProjectileState {
		Init,					//Default State	- Projectile is first created.
		Activate,				//Set State		- Projectile is activated.
		Deactivate,				//Set State		- Projectile is deactivated.
		Active_Processing,		//Access State	- Projectile is becoming Active.
		Active_Resting,			//Access State	- Projectile is active, but resting.
		Active_Normal,			//Access State	- Projectile is following a normal pattern.
		Active_NormalSplit,		//Access State	- Projectile is following a normal pattern while splitting.
		Active_Homing,			//Access State	- Projectile is moving towards another object.
		Active_HomingSplit,		//Access State	- Projectile is moving towards another object while splitting.
		Inactive,				//Access State	- Projectile is Inactive.
	}
	
	///////////////////////////////////////////////////////////////////////////////////////////////
	// INSTANCE VARIABLES
	///////////////////////////////////////////////////////////////////////////////////////////////
	public float lifespan;
	public float accelleration;
	public float turning;
	public float rest;
	public boolean faceDirection;
	protected float speed;
	protected float direction;
	protected float deltaX;
	protected float deltaY;
	protected int timesSplit;
	protected IntervalSplitter stepSplitter;
	protected InstanceSplitter createSplitter;
	protected InstanceSplitter destroySplitter;
	protected Object target;
	protected ProjectileState state;
	
	//For BulletHandler
	public int index;
	protected BulletHandler handler;
	
	///////////////////////////////////////////////////////////////////////////////////////////////
	// CONSTRUCTORS
	///////////////////////////////////////////////////////////////////////////////////////////////
	public Projectile clone() {
		return new Projectile(this);
	}
	
	public Projectile(Projectile p) {
		super(p);
		
		//Projectile
		this.lifespan = p.lifespan;
		this.accelleration = p.accelleration;
		this.turning = p.turning;
		this.rest = p.rest;
		this.faceDirection = p.faceDirection;
		this.speed = p.speed;
		this.direction = p.direction;
		this.deltaX = p.deltaX;
		this.deltaY = p.deltaY;
		this.timesSplit = p.timesSplit;
		this.setSplitter(SplitterType.Create, Splitter.duplicate(p.createSplitter));
		this.setSplitter(SplitterType.Step, Splitter.duplicate(p.stepSplitter));
		this.setSplitter(SplitterType.Destroy, Splitter.duplicate(p.destroySplitter));
		this.target = p.target;
		this.setState(p.state);
		
		//Handled by BulletHandler
		this.index = 0;
		this.handler = null;
	}
	
	public Projectile() {
		this(Projectile.DefaultName);
	}
	
	public Projectile(String name) {
		super(name);
		
		//Projectile
		this.lifespan = -1.0f;
		this.accelleration = 0.0f;
		this.turning = 0.0f;
		this.rest = 0.0f;
		this.faceDirection = false;
		this.speed = 0.0f;
		this.direction = 0.0f;
		this.deltaX = 0.0f;
		this.deltaY = 0.0f;
		this.timesSplit = 0;
		this.setSplitter(SplitterType.Create, null);
		this.setSplitter(SplitterType.Step, null);
		this.setSplitter(SplitterType.Destroy, null);
		this.target = null;
		this.state = ProjectileState.Init;
		
		//Handled by BulletHandler
		this.index = 0;
		this.handler = null;
	}
	
	public Projectile(BulletHandler h, int index) {
		this();
		this.handler = h;
		this.index = index;
		this.state = ProjectileState.Inactive;
	}
	
	public Projectile(TextureRegion region) {
		this();
		setRegion(region);
	}
	
	public Projectile(TextureRegion region, float direction, float speed, boolean faceDirection) {
		this(region);
		this.faceDirection = faceDirection;
		setVelocity(direction, speed);
	}
	
	///////////////////////////////////////////////////////////////////////////////////////////////
	// MUTATORS / SETTERS
	///////////////////////////////////////////////////////////////////////////////////////////////
	public void set(Projectile p) {
		super.set(p);
		//Projectile
		this.lifespan = p.lifespan;
		this.accelleration = p.accelleration;
		this.turning = p.turning;
		this.rest = p.rest;
		this.faceDirection = p.faceDirection;
		this.speed = p.speed;
		this.direction = p.direction;
		this.deltaX = p.deltaX;
		this.deltaY = p.deltaY;
		this.timesSplit = p.timesSplit;
		this.setSplitter(SplitterType.Create, Splitter.duplicate(p.createSplitter));
		this.setSplitter(SplitterType.Step, Splitter.duplicate(p.stepSplitter));
		this.setSplitter(SplitterType.Destroy, Splitter.duplicate(p.destroySplitter));
		this.target = p.target;
		this.setState(p.state);
	}
	
	private void setState(ProjectileState s) {
		if(state == s)
			return;
		switch(s) {
		case Init:
			//Does nothing, simply a starting state.
			break;
		case Activate:
			//Does nothing, waits one logical step until completely activating.
			break;
		case Deactivate:
			setState(ProjectileState.Inactive);
			return;
		case Active_Resting:
			break;
		case Active_Processing:
			if(rest > 0.0f)
				setState(ProjectileState.Active_Resting);
			else if(target != null)
				setState(ProjectileState.Active_Homing);
			else
				setState(ProjectileState.Active_Normal);
			return;
		case Active_Normal:
			if(stepSplitter != null)
				s = ProjectileState.Active_NormalSplit;
			break;
		case Active_NormalSplit:
			break;
		case Active_Homing:
			if(stepSplitter != null)
				s = ProjectileState.Active_HomingSplit;
			break;
		case Active_HomingSplit:
			break;
		case Inactive:
			break;
		}
		state = s;
	}
	
	public void setSplitter(SplitterType type, Splitter s) {
		switch(type) {
		case Create:
			createSplitter = (InstanceSplitter)s;
			break;
		case Step:
			stepSplitter = (IntervalSplitter)s;
			break;
		case Destroy:
			destroySplitter = (InstanceSplitter)s;
			break;
		}
	}
	
	public void setSpeed(float speed) {
		setVelocity(direction, speed);
	}
	
	public void setDirection(float direction) {
		setVelocity(direction, speed);
	}
	
	public void setMotionX(float dx) {
		setMotion(dx, deltaY);
	}
	
	public void setMotionY(float dy) {
		setMotion(deltaX, dy);
	}
	
	public void setMotion(float dx, float dy) {
		this.deltaX = dx;
		this.deltaY = dy;
		refactorMotion();
	}
	
	public void setVelocity(float direction, float speed) {
		if(speed < 0.0f) {
			if(accelleration < 0.0f)
				accelleration *= -1.0f;
			this.direction = Helper.ReverseAngle(direction);
			this.speed = -speed;
		} else {
			this.direction = Helper.NormalizeAngle(direction);
			this.speed = speed;
		}
		refactorVelocity();
	}
	
	public void setIndex(int i) {
		this.index = i;
	}
	
	///////////////////////////////////////////////////////////////////////////////////////////////
	// ACCESSORS / GETTERS
	///////////////////////////////////////////////////////////////////////////////////////////////
	
	public float getSpeed() {
		return speed;
	}
	
	public float getDirection() {
		return direction;
	}
	
	public float getDeltaX() {
		return deltaX;
	}
	
	public float getDeltaY() {
		return deltaY;
	}
	
	public Splitter getSplitter(SplitterType type) {
		switch(type) {
		case Create:
			return createSplitter;
		case Step:
			return stepSplitter;
		case Destroy:
			return destroySplitter;
		}
		
		//To please compiler
		return null;
	}
	
	///////////////////////////////////////////////////////////////////////////////////////////////
	// OVERRIDES
	///////////////////////////////////////////////////////////////////////////////////////////////
	
	@Override
	public void create() {
		if(createSplitter != null)
			createSplitter.actuate(this);
	}
	
	@Override
	public void destroy() {
		if(destroySplitter != null)
			destroySplitter.actuate(this);
		deactivate();
	}
	
	@Override
	public void act(float dt) {
		stateTime += dt;
		
		switch(state) {
		case Activate:
			create();
			setState(ProjectileState.Active_Processing);
			break;
		case Active_Resting:
			updateRest(dt);
			break;
		case Active_Homing:
			turnTowardsTarget();
			applyStandardLogic(dt);
			break;
		case Active_HomingSplit:
			turnTowardsTarget();
			applyStandardLogic(dt);
			updateStepSplitter(dt);
			break;
		case Active_Normal:
			applyStandardLogic(dt);
			break;
		case Active_NormalSplit:
			applyStandardLogic(dt);
			updateStepSplitter(dt);
			break;
		default:
			//Bullet is inactive or init, should not have logic.
			deactivate();
			break;
		}
	}

	@Override
	public void endOfAnimation() {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void checkCollisionWithObject(AbstractObject other) {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void checkCollisionWithObjects(ObjectHandler handler) {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void collisionHasOccurredWithObject(AbstractObject other) {
		deactivate();
	}
	

	@Override
	public void checkCollisionWithProjectiles(BulletHandler handler) {
	}
	
	@Override
	public int hashCode() {
		int hash = Integer.MIN_VALUE;
		hash += (accelleration < 0.0f) ? 31 * (Math.abs(accelleration) * Projectile.SigFig) : 16 * (accelleration * Projectile.SigFig);
		hash += (turning < 0.0f) ? 15 * (Math.abs(turning) * Projectile.SigFig) : 7 * (turning * Projectile.SigFig);
		hash += 3 * (rest * Projectile.SigFig);
		hash += 11 * (speed * Projectile.SigFig);
		hash += 5 * (direction * Projectile.SigFig);
		return hash;
	}
	
	///////////////////////////////////////////////////////////////////////////////////////////////
	// PUBLIC METHODS
	///////////////////////////////////////////////////////////////////////////////////////////////
	public void activate() {
		switch(state) {
		case Init:
		case Inactive:
		case Deactivate:
			setState(ProjectileState.Activate);
			break;
		default:
			break;
		}
	}
	
	public void deactivate() {
		switch(state) {
		case Init:
		case Inactive:
		case Deactivate:
			break;
		default:
			setState(ProjectileState.Deactivate);
			if(handler.getActiveProjectiles() > index)
				handler.removeProjectile(index);
			break;
		}
	}
	
	public void moveTowardsObject(Object other, float speed) {
		setVelocity(MathUtils.atan2(other.y - y, other.x - x), speed);
	}
	
	public void followObject(Object target, float speed) {
		followObject(target, MathUtils.PI * 2, speed);
	}
	
	public void followObject(Object target, float turningRadius, float speed) {
		this.target = target;
		this.turning = turningRadius;
		this.setSpeed(speed);
	}

	public void influenceProjectile(Projectile projectile) {
		projectile.setPosition(projectile.x + this.x, projectile.y + this.y);
		projectile.setDirection(projectile.direction + this.direction);
	}
	
	///////////////////////////////////////////////////////////////////////////////////////////////
	// PRIVATE METHODS
	///////////////////////////////////////////////////////////////////////////////////////////////
	private void updateRest(float dt) {
		if(rest > 0.0f) {
			rest -= dt;
			if(rest < 0.0f) {
				rest = 0.0f;
				setState(ProjectileState.Active_Processing);
			}
		}
	}
	
	private void turnTowardsTarget() {
		float shouldFace = Helper.NormalizeAngle(MathUtils.atan2(target.y - y, target.x - x));
		float difference = Helper.NormalizeAngle(direction - shouldFace);
		if(difference > MathUtils.PI) {
			if(turning < 0.0f)
				turning *= -1.0f;
		} else if(turning > 0.0f)
			turning *= -1.0f;
	}
	
	private void applyStandardLogic(float dt) {
		//Translations
		translate(deltaX * dt, deltaY * dt);
		if(turning != 0.0f || accelleration != 0.0f)
			setVelocity(direction + (turning * dt), speed + (accelleration * dt));
		
		//Face Direction
		if(faceDirection)
			rotation = direction;
		
		//Check lifespan
		if(lifespan > 0.0f) {
			lifespan -= dt;
			if(lifespan <= 0.0f)
				destroy();
		}
	}
	
	private void updateStepSplitter(float dt) {
		if(stepSplitter.totalSplits < 0 || (stateTime >= stepSplitter.timeUntilSplits && timesSplit < stepSplitter.totalSplits )) {
			float nextSplitTime = stepSplitter.timeUntilSplits + (stepSplitter.timeBetweenSplits * timesSplit);
			if(stateTime >= nextSplitTime) {
				timesSplit++;
				stepSplitter.actuate(this);
			}
		}
	}
	
	private void refactorVelocity() {
		deltaX = MathUtils.cos(direction) * speed;
		deltaY = MathUtils.sin(direction) * speed;
	}
	
	private void refactorMotion() {
		direction = Helper.NormalizeAngle(MathUtils.atan2(deltaY, deltaX));
		speed = (float)Math.sqrt((double)((deltaX * deltaX) + (deltaY * deltaY)));
	}
	
	///////////////////////////////////////////////////////////////////////////////////////////////
	// PUBLIC STATIC METHODS
	///////////////////////////////////////////////////////////////////////////////////////////////
	public static Projectile[] SplitByTwo(Projectile template) {
		return SplitByTwo(template, 0.0f);
	}
	
	public static Projectile[] SplitByTwo(Projectile template, float inset) {
		Projectile l[] = new Projectile[2];
		float angle = MathUtils.PI / 2;
		l[0] = template.clone();
		l[1] = template.clone();
		l[0].setDirection(angle - inset);
		l[1].setDirection(-angle + inset);
		return l;
	}
	
	public static Projectile[] StarList(Projectile template, int num) {
		return StarList(template, num, 0.0f);
	}
	
	public static Projectile[] StarList(Projectile template, int num, float offset) {
		Projectile l[] = new Projectile[num];
		for(int i = 0; i < num; i++) {
			l[i] = template.clone();
			l[i].setDirection(offset + ((i * Helper.PI2) / num));
		}
		return l;
	}
}
